# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# pyre-strict

from libcst.codemod import CodemodTest
from libcst.codemod.commands.rename_typing_generic_aliases import (
    RenameTypingGenericAliases,
)


class TestRenameCommand(CodemodTest):
    TRANSFORM = RenameTypingGenericAliases

    def test_rename_typing_generic_alias(self) -> None:
        before = """
            from typing import List, Set, Dict, FrozenSet, Tuple
            x: List[int] = []
            y: Set[int] = set()
            z: Dict[str, int] = {}
            a: FrozenSet[str] = frozenset()
            b: Tuple[int, str] = (1, "hello")
        """
        after = """
            x: list[int] = []
            y: set[int] = set()
            z: dict[str, int] = {}
            a: frozenset[str] = frozenset()
            b: tuple[int, str] = (1, "hello")
        """
        self.assertCodemod(before, after)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# pyre-strict

from libcst.codemod import CodemodTest
from libcst.codemod.commands.convert_union_to_or import ConvertUnionToOrCommand


class TestConvertUnionToOrCommand(CodemodTest):
    TRANSFORM = ConvertUnionToOrCommand

    def test_simple_union(self) -> None:
        before = """
            from typing import Union
            x: Union[int, str]
        """
        after = """
            x: int | str
        """
        self.assertCodemod(before, after)

    def test_nested_union(self) -> None:
        before = """
            from typing import Union
            x: Union[int, Union[str, float]]
        """
        after = """
            x: int | str | float
        """
        self.assertCodemod(before, after)

    def test_single_type_union(self) -> None:
        before = """
            from typing import Union
            x: Union[int]
        """
        after = """
            x: int
        """
        self.assertCodemod(before, after)

    def test_union_with_alias(self) -> None:
        before = """
            import typing as t
            x: t.Union[int, str]
        """
        after = """
            import typing as t
            x: int | str
        """
        self.assertCodemod(before, after)

    def test_union_with_unused_import(self) -> None:
        before = """
            from typing import Union, List
            x: Union[int, str]
        """
        after = """
            from typing import List
            x: int | str
        """
        self.assertCodemod(before, after)

    def test_union_no_import(self) -> None:
        before = """
            x: Union[int, str]
        """
        after = """
            x: Union[int, str]
        """
        self.assertCodemod(before, after)

    def test_union_in_function(self) -> None:
        before = """
            from typing import Union
            def foo(x: Union[int, str]) -> Union[float, None]:
                ...
        """
        after = """
            def foo(x: int | str) -> float | None:
                ...
        """
        self.assertCodemod(before, after)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from pathlib import Path
from unittest import TestCase

from libcst import CSTTransformer, parse_module
from libcst._parser.entrypoints import is_native

fixtures: Path = Path(__file__).parent.parent.parent / "native/libcst/tests/fixtures"


class NOOPTransformer(CSTTransformer):
    pass


class RoundTripTests(TestCase):
    def _get_fixtures(self) -> list[Path]:
        if not is_native():
            self.skipTest("pure python parser doesn't work with this")
        self.assertTrue(fixtures.exists(), f"{fixtures} should exist")
        files = list(fixtures.iterdir())
        self.assertGreater(len(files), 0)
        return files

    def test_clean_roundtrip(self) -> None:
        for file in self._get_fixtures():
            with self.subTest(file=str(file)):
                src = file.read_text(encoding="utf-8")
                mod = parse_module(src)
                self.maxDiff = None
                self.assertEqual(mod.code, src)

    def test_transform_roundtrip(self) -> None:
        transformer = NOOPTransformer()
        self.maxDiff = None
        for file in self._get_fixtures():
            with self.subTest(file=str(file)):
                src = file.read_text(encoding="utf-8")
                mod = parse_module(src)
                new_mod = mod.visit(transformer)
                self.assertEqual(src, new_mod.code)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Optional
from unittest.mock import Mock

import libcst as cst
from libcst.metadata import (
    BatchableMetadataProvider,
    MetadataWrapper,
    VisitorMetadataProvider,
)
from libcst.testing.utils import UnitTest


class MetadataWrapperTest(UnitTest):
    def test_copies_tree(self) -> None:
        m = cst.parse_module("pass")
        mw = MetadataWrapper(m)
        self.assertTrue(mw.module.deep_equals(m))
        self.assertIsNot(mw.module, m)

    def test_unsafe_skip_copy(self) -> None:
        m = cst.parse_module("pass")
        mw = MetadataWrapper(m, unsafe_skip_copy=True)
        self.assertIs(mw.module, m)

    def test_equality_by_identity(self) -> None:
        m = cst.parse_module("pass")
        mw1 = MetadataWrapper(m)
        mw2 = MetadataWrapper(m)
        self.assertEqual(mw1, mw1)
        self.assertEqual(mw2, mw2)
        self.assertNotEqual(mw1, mw2)

    def test_hash_by_identity(self) -> None:
        m = cst.parse_module("pass")
        mw1 = MetadataWrapper(m)
        mw2 = MetadataWrapper(m, unsafe_skip_copy=True)
        mw3 = MetadataWrapper(m, unsafe_skip_copy=True)
        self.assertEqual(hash(mw1), hash(mw1))
        self.assertEqual(hash(mw2), hash(mw2))
        self.assertEqual(hash(mw3), hash(mw3))
        self.assertNotEqual(hash(mw1), hash(mw2))
        self.assertNotEqual(hash(mw1), hash(mw3))
        self.assertNotEqual(hash(mw2), hash(mw3))

    @staticmethod
    def ignore_args(*args: object, **kwargs: object) -> tuple[object, ...]:
        return (args, kwargs)

    def test_metadata_cache(self) -> None:
        class DummyMetadataProvider(BatchableMetadataProvider[None]):
            gen_cache = self.ignore_args

        m = cst.parse_module("pass")
        mw = MetadataWrapper(m)
        with self.assertRaisesRegex(
            Exception, "Cache is required for initializing DummyMetadataProvider."
        ):
            mw.resolve(DummyMetadataProvider)

        class SimpleCacheMetadataProvider(BatchableMetadataProvider[object]):
            gen_cache = self.ignore_args

            def __init__(self, cache: object) -> None:
                super().__init__(cache)
                self.cache = cache

            def visit_Pass(self, node: cst.Pass) -> Optional[bool]:
                self.set_metadata(node, self.cache)

        cached_data = object()
        mw = MetadataWrapper(m, cache={SimpleCacheMetadataProvider: cached_data})
        pass_node = cst.ensure_type(mw.module.body[0], cst.SimpleStatementLine).body[0]
        self.assertEqual(
            mw.resolve(SimpleCacheMetadataProvider)[pass_node], cached_data
        )

    def test_resolve_provider_twice(self) -> None:
        """
        Tests that resolving the same provider twice is a no-op
        """
        mock = Mock()

        class ProviderA(VisitorMetadataProvider[bool]):
            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_a()

        module = cst.parse_module("pass")
        wrapper = MetadataWrapper(module)

        wrapper.resolve(ProviderA)
        mock.visited_a.assert_called_once()

        wrapper.resolve(ProviderA)
        mock.visited_a.assert_called_once()

    def test_resolve_dependent_provider_twice(self) -> None:
        """
        Tests that resolving the same provider twice is a no-op
        """
        mock = Mock()

        class ProviderA(VisitorMetadataProvider[bool]):
            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_a()

        class ProviderB(VisitorMetadataProvider[bool]):
            METADATA_DEPENDENCIES = (ProviderA,)

            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_b()

        module = cst.parse_module("pass")
        wrapper = MetadataWrapper(module)

        wrapper.resolve(ProviderA)
        mock.visited_a.assert_called_once()

        wrapper.resolve(ProviderB)
        mock.visited_a.assert_called_once()
        mock.visited_b.assert_called_once()

        wrapper.resolve(ProviderA)
        mock.visited_a.assert_called_once()
        mock.visited_b.assert_called_once()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING

from libcst import parse_module
from libcst.display import dump_graphviz
from libcst.testing.utils import UnitTest

if TYPE_CHECKING:
    from libcst import Module


class CSTDumpGraphvizTest(UnitTest):
    """Check dump_graphviz contains CST nodes."""

    source_code: str = dedent(
        r"""
        def foo(a: str) -> None:
            pass ;
            pass
            return
        """[
            1:
        ]
    )
    cst: Module

    @classmethod
    def setUpClass(cls) -> None:
        cls.cst = parse_module(cls.source_code)

    def _assert_node(self, node_name: str, graphviz_str: str) -> None:
        self.assertIn(
            node_name, graphviz_str, f"No node {node_name} found in graphviz_dump"
        )

    def _check_essential_nodes_in_tree(self, graphviz_str: str) -> None:
        # Check CST nodes are present in graphviz string
        self._assert_node("Module", graphviz_str)
        self._assert_node("FunctionDef", graphviz_str)
        self._assert_node("Name", graphviz_str)
        self._assert_node("Parameters", graphviz_str)
        self._assert_node("Param", graphviz_str)
        self._assert_node("Annotation", graphviz_str)
        self._assert_node("IndentedBlock", graphviz_str)
        self._assert_node("SimpleStatementLine", graphviz_str)
        self._assert_node("Pass", graphviz_str)
        self._assert_node("Return", graphviz_str)

        # Check CST values are present in graphviz string
        self._assert_node("<foo>", graphviz_str)
        self._assert_node("<a>", graphviz_str)
        self._assert_node("<str>", graphviz_str)
        self._assert_node("<None>", graphviz_str)

    def test_essential_tree(self) -> None:
        """Check essential nodes are present in the CST graphviz dump."""
        graphviz_str = dump_graphviz(self.cst)
        self._check_essential_nodes_in_tree(graphviz_str)

    def test_full_tree(self) -> None:
        """Check all nodes are present in the CST graphviz dump."""
        graphviz_str = dump_graphviz(
            self.cst,
            show_whitespace=True,
            show_defaults=True,
            show_syntax=True,
        )
        self._check_essential_nodes_in_tree(graphviz_str)

        self._assert_node("Semicolon", graphviz_str)
        self._assert_node("SimpleWhitespace", graphviz_str)
        self._assert_node("Newline", graphviz_str)
        self._assert_node("TrailingWhitespace", graphviz_str)

        self._assert_node("<>", graphviz_str)
        self._assert_node("< >", graphviz_str)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unittest import TestCase

from libcst import (
    Annotation,
    CSTNode,
    FunctionDef,
    IndentedBlock,
    Module,
    Param,
    parse_module,
    Pass,
    Semicolon,
    SimpleStatementLine,
)

from libcst.helpers import (
    get_node_fields,
    is_default_node_field,
    is_syntax_node_field,
    is_whitespace_node_field,
)


class _NodeFieldsTest(TestCase):
    """Node fields related tests."""

    module: Module
    annotation: Annotation
    param: Param
    _pass: Pass
    semicolon: Semicolon
    statement: SimpleStatementLine
    indent: IndentedBlock
    function: FunctionDef

    @classmethod
    def setUpClass(cls) -> None:
        """Parse a simple CST and references interesting nodes."""
        cls.module = parse_module(
            "def foo(a: str) -> None:\n    pass ; pass\n    return\n"
        )
        # /!\ Direct access to nodes
        # This is done for test purposes on a known CST
        # -> For "real code", use visitors to do this "the correct way"

        # pyre-ignore[8]: direct access for tests
        cls.function = cls.module.body[0]
        cls.param = cls.function.params.params[0]
        # pyre-ignore[8]: direct access for tests
        cls.annotation = cls.param.annotation
        # pyre-ignore[8]: direct access for tests
        cls.indent = cls.function.body
        # pyre-ignore[8]: direct access for tests
        cls.statement = cls.indent.body[0]
        # pyre-ignore[8]: direct access for tests
        cls._pass = cls.statement.body[0]
        # pyre-ignore[8]: direct access for tests
        cls.semicolon = cls.statement.body[0].semicolon

    def test__cst_correctness(self) -> None:
        """Test that the CST is correctly parsed."""
        self.assertIsInstance(self.module, Module)
        self.assertIsInstance(self.annotation, Annotation)
        self.assertIsInstance(self.param, Param)
        self.assertIsInstance(self._pass, Pass)
        self.assertIsInstance(self.semicolon, Semicolon)
        self.assertIsInstance(self.statement, SimpleStatementLine)
        self.assertIsInstance(self.indent, IndentedBlock)
        self.assertIsInstance(self.function, FunctionDef)


class IsWhitespaceNodeFieldTest(_NodeFieldsTest):
    """``is_whitespace_node_field`` tests."""

    def _check_fields(self, is_filtered_field: dict[str, bool], node: CSTNode) -> None:
        fields = get_node_fields(node)
        self.assertEqual(len(is_filtered_field), len(fields))
        for field in fields:
            self.assertEqual(
                is_filtered_field[field.name],
                is_whitespace_node_field(node, field),
                f"Node ``{node.__class__.__qualname__}`` field '{field.name}' "
                f"{'should have' if is_filtered_field[field.name] else 'should not have'} "
                "been filtered by ``is_whitespace_node_field``",
            )

    def test_module(self) -> None:
        """Check if a CST Module node is correctly filtered."""
        is_filtered_field = {
            "body": False,
            "header": True,
            "footer": True,
            "encoding": False,
            "default_indent": False,
            "default_newline": False,
            "has_trailing_newline": False,
        }
        self._check_fields(is_filtered_field, self.module)

    def test_annotation(self) -> None:
        """Check if a CST Annotation node is correctly filtered."""
        is_filtered_field = {
            "annotation": False,
            "whitespace_before_indicator": True,
            "whitespace_after_indicator": True,
        }
        self._check_fields(is_filtered_field, self.annotation)

    def test_param(self) -> None:
        """Check if a CST Param node is correctly filtered."""
        is_filtered_field = {
            "name": False,
            "annotation": False,
            "equal": False,
            "default": False,
            "comma": False,
            "star": False,
            "whitespace_after_star": True,
            "whitespace_after_param": True,
        }
        self._check_fields(is_filtered_field, self.param)

    def test_semicolon(self) -> None:
        """Check if a CST Semicolon node is correctly filtered."""
        is_filtered_field = {
            "whitespace_before": True,
            "whitespace_after": True,
        }
        self._check_fields(is_filtered_field, self.semicolon)

    def test_statement(self) -> None:
        """Check if a CST SimpleStatementLine node is correctly filtered."""
        is_filtered_field = {
            "body": False,
            "leading_lines": True,
            "trailing_whitespace": True,
        }
        self._check_fields(is_filtered_field, self.statement)

    def test_indent(self) -> None:
        """Check if a CST IndentedBlock node is correctly filtered."""
        is_filtered_field = {
            "body": False,
            "header": True,
            "indent": True,
            "footer": True,
        }
        self._check_fields(is_filtered_field, self.indent)

    def test_function(self) -> None:
        """Check if a CST FunctionDef node is correctly filtered."""
        is_filtered_field = {
            "name": False,
            "params": False,
            "body": False,
            "decorators": False,
            "returns": False,
            "asynchronous": False,
            "leading_lines": True,
            "lines_after_decorators": True,
            "whitespace_after_def": True,
            "whitespace_after_name": True,
            "whitespace_before_params": True,
            "whitespace_before_colon": True,
            "type_parameters": False,
            "whitespace_after_type_parameters": True,
        }
        self._check_fields(is_filtered_field, self.function)


class IsSyntaxNodeFieldTest(_NodeFieldsTest):
    """``is_syntax_node_field`` tests."""

    def _check_fields(self, is_filtered_field: dict[str, bool], node: CSTNode) -> None:
        fields = get_node_fields(node)
        self.assertEqual(len(is_filtered_field), len(fields))
        for field in fields:
            self.assertEqual(
                is_filtered_field[field.name],
                is_syntax_node_field(node, field),
                f"Node ``{node.__class__.__qualname__}`` field '{field.name}' "
                f"{'should have' if is_filtered_field[field.name] else 'should not have'} "
                "been filtered by ``is_syntax_node_field``",
            )

    def test_module(self) -> None:
        """Check if a CST Module node is correctly filtered."""
        is_filtered_field = {
            "body": False,
            "header": False,
            "footer": False,
            "encoding": True,
            "default_indent": True,
            "default_newline": True,
            "has_trailing_newline": True,
        }
        self._check_fields(is_filtered_field, self.module)

    def test_param(self) -> None:
        """Check if a CST Param node is correctly filtered."""
        is_filtered_field = {
            "name": False,
            "annotation": False,
            "equal": True,
            "default": False,
            "comma": True,
            "star": False,
            "whitespace_after_star": False,
            "whitespace_after_param": False,
        }
        self._check_fields(is_filtered_field, self.param)

    def test_pass(self) -> None:
        """Check if a CST Pass node is correctly filtered."""
        is_filtered_field = {
            "semicolon": True,
        }
        self._check_fields(is_filtered_field, self._pass)


class IsDefaultNodeFieldTest(_NodeFieldsTest):
    """``is_default_node_field`` tests."""

    def _check_fields(self, is_filtered_field: dict[str, bool], node: CSTNode) -> None:
        fields = get_node_fields(node)
        self.assertEqual(len(is_filtered_field), len(fields))
        for field in fields:
            self.assertEqual(
                is_filtered_field[field.name],
                is_default_node_field(node, field),
                f"Node ``{node.__class__.__qualname__}`` field '{field.name}' "
                f"{'should have' if is_filtered_field[field.name] else 'should not have'} "
                "been filtered by ``is_default_node_field``",
            )

    def test_module(self) -> None:
        """Check if a CST Module node is correctly filtered."""
        is_filtered_field = {
            "body": False,
            "header": True,
            "footer": True,
            "encoding": True,
            "default_indent": True,
            "default_newline": True,
            "has_trailing_newline": True,
        }
        self._check_fields(is_filtered_field, self.module)

    def test_annotation(self) -> None:
        """Check if a CST Annotation node is correctly filtered."""
        is_filtered_field = {
            "annotation": False,
            "whitespace_before_indicator": False,
            "whitespace_after_indicator": True,
        }
        self._check_fields(is_filtered_field, self.annotation)

    def test_param(self) -> None:
        """Check if a CST Param node is correctly filtered."""
        is_filtered_field = {
            "name": False,
            "annotation": False,
            "equal": True,
            "default": True,
            "comma": True,
            "star": False,
            "whitespace_after_star": True,
            "whitespace_after_param": True,
        }
        self._check_fields(is_filtered_field, self.param)

    def test_statement(self) -> None:
        """Check if a CST SimpleStatementLine node is correctly filtered."""
        is_filtered_field = {
            "body": False,
            "leading_lines": True,
            "trailing_whitespace": True,
        }
        self._check_fields(is_filtered_field, self.statement)

    def test_indent(self) -> None:
        """Check if a CST IndentedBlock node is correctly filtered."""
        is_filtered_field = {
            "body": False,
            "header": True,
            "indent": True,
            "footer": True,
        }
        self._check_fields(is_filtered_field, self.indent)

    def test_function(self) -> None:
        """Check if a CST FunctionDef node is correctly filtered."""
        is_filtered_field = {
            "name": False,
            "params": False,
            "body": False,
            "decorators": True,
            "returns": False,
            "asynchronous": True,
            "leading_lines": True,
            "lines_after_decorators": True,
            "whitespace_after_def": True,
            "whitespace_after_name": True,
            "whitespace_before_params": True,
            "whitespace_before_colon": True,
            "type_parameters": True,
            "whitespace_after_type_parameters": True,
        }
        self._check_fields(is_filtered_field, self.function)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# pyre-strict

from libcst.codemod import CodemodTest
from libcst.codemod.commands.fix_variadic_callable import FixVariadicCallableCommmand


class TestFixVariadicCallableCommmand(CodemodTest):
    TRANSFORM = FixVariadicCallableCommmand

    def test_callable_typing(self) -> None:
        before = """
            from typing import Callable
            x: Callable[[...], int] = ...
        """
        after = """
            from typing import Callable
            x: Callable[..., int] = ...
        """
        self.assertCodemod(before, after)

    def test_callable_typing_alias(self) -> None:
        before = """
            import typing as t
            x: t.Callable[[...], int] = ...
        """
        after = """
            import typing as t
            x: t.Callable[..., int] = ...
        """
        self.assertCodemod(before, after)

    def test_callable_import_alias(self) -> None:
        before = """
            from typing import Callable as C
            x: C[[...], int] = ...
        """
        after = """
            from typing import Callable as C
            x: C[..., int] = ...
        """
        self.assertCodemod(before, after)

    def test_callable_with_optional(self) -> None:
        before = """
            from typing import Callable
            def foo(bar: Optional[Callable[[...], int]]) -> Callable[[...], int]:
                ...
        """
        after = """
            from typing import Callable
            def foo(bar: Optional[Callable[..., int]]) -> Callable[..., int]:
                ...
        """
        self.assertCodemod(before, after)

    def test_callable_with_arguments(self) -> None:
        before = """
            from typing import Callable
            x: Callable[[int], int]
        """
        after = """
            from typing import Callable
            x: Callable[[int], int]
        """
        self.assertCodemod(before, after)

    def test_callable_with_variadic_arguments(self) -> None:
        before = """
            from typing import Callable
            x: Callable[[int, int, ...], int]
        """
        after = """
            from typing import Callable
            x: Callable[[int, int, ...], int]
        """
        self.assertCodemod(before, after)

    def test_callable_no_arguments(self) -> None:
        before = """
            from typing import Callable
            x: Callable
        """
        after = """
            from typing import Callable
            x: Callable
        """
        self.assertCodemod(before, after)
